%%%-------------------------------------------------------------------
%%% File    : irc.erl
%%% Author  : Samuel Tardieu <sam@rfc1149.net>
%%% Description : IRC helper functions
%%%
%%% Created : 4 June 2007 by Samuel Tardieu <sam@rfc1149.net>
%%%-------------------------------------------------------------------
-module(irc).

%% API
-export([format/2,
	 privmsg/2, notice/2, ctcp/2, ctcp/3, ctcp_reply/3,
	 parse/1, parse_line/1, parse_cmd/1, nick/1, user/1, host/1, 
	 userhost/1]).

%%====================================================================
%% API
%%====================================================================
%%--------------------------------------------------------------------
%% Function: format/2
%% Description: Return a properly formatted IRC message
%%--------------------------------------------------------------------
format(Command, []) -> [Command, "\r\n"];
format(Command, [Arg]) -> [Command, " :", Arg, "\r\n"];
format(Command, Args) ->
    [Last | Rev] = lists:reverse(Args),
    [Command, $ , join(lists:reverse(Rev), $ ), " :", Last, "\r\n"].

%%--------------------------------------------------------------------
%% Function: privmsg/2
%% Description: Return a properly formatted PRIVMSG message
%%--------------------------------------------------------------------
privmsg(Target, Data) -> format("PRIVMSG", [Target, Data]).

%%--------------------------------------------------------------------
%% Function: notice/2
%% Description: Return a properly formatted PRIVMSG message
%%--------------------------------------------------------------------
notice(Target, Data) -> format("NOTICE", [Target, Data]).

%%--------------------------------------------------------------------
%% Function: ctcp/2
%% Description: Return a properly formatted CTCP query message
%%--------------------------------------------------------------------
ctcp(Target, Action) -> privmsg(Target, [1, Action, 1]).

%%--------------------------------------------------------------------
%% Function: ctcp/3
%% Description: Return a properly formatted CTCP query message
%%--------------------------------------------------------------------
ctcp(Target, Action, Data) -> ctcp(Target, [Action, $ , Data]).

%%--------------------------------------------------------------------
%% Function: ctcp_reply/3
%% Description: Return a properly formatted CTCP reply message
%%--------------------------------------------------------------------
ctcp_reply(Target, Action, Data) -> notice(Target, [1, Action, $ , Data, 1]).

%%--------------------------------------------------------------------
%% Function: parse/1
%% Description: Parse a IRC string and return {Sender, Action, Args}
%%              If there are no sender, undefined is used instead
%%--------------------------------------------------------------------
parse([$: | Str]) ->
    [Sender, Action | Args] = parse_tokens(Str),
    {Sender, Action, Args};
parse(Str) ->
    [Action | Args] = parse_tokens(Str),
    {undefined, Action, Args}.

%%--------------------------------------------------------------------
%% Function: parse_line/1
%% Description: Parse a IRC string and return {Sender, Action, Args}
%%              If there are no sender, undefined is used instead.
%%              The two last characters are removed (\r\n)
%%--------------------------------------------------------------------
parse_line(Str) ->
    [_, _ | Rev] = lists:reverse(Str),
    parse(lists:reverse(Rev)).

%%--------------------------------------------------------------------
%% Function : parse_cmd/1
%% Description : Parse an TCP string message and return a list which
%%		 contains {Sender, Action, Args} for each command
%%		 of the TCP string message
%%--------------------------------------------------------------------
parse_cmd(Str) ->
   case string:str(Str, "\n") of
	0 -> [];
	I -> Toto = parse_line(string:substr(Str,1,I)),
	     Titi = irc:parse_cmd(string:substr(Str,I+1)),
	     [Toto|Titi]
   end.


%%--------------------------------------------------------------------
%% Function: nick/1
%% Description: Extract nick from nick!server@host
%%--------------------------------------------------------------------
nick(Str) ->
    hd(string:tokens(Str, "!")).

%%--------------------------------------------------------------------
%% Function: user/1
%% Description: Extract user from nick!server@host
%%--------------------------------------------------------------------
user(Str) ->
    hd(string:tokens(userhost(Str), "@")).

%%--------------------------------------------------------------------
%% Function: host/1
%% Description: Extract user from nick!server@host
%%--------------------------------------------------------------------
host(Str) ->
    snd(string:tokens(userhost(Str), "@")).

%%--------------------------------------------------------------------
%% Function: userhost/1
%% Description: Extract user from nick!server@host
%%--------------------------------------------------------------------
userhost(Str) ->
    snd(string:tokens(Str, "!")).

%%====================================================================
%% Internal functions
%%====================================================================
join([], _) -> [];
join([A], _) -> [A];
join([H | T], Sep) -> [H, Sep, join(T, Sep)].

parse_tokens([$: | Rest]) -> [Rest];    
parse_tokens(Str) ->
    case string:chr(Str, $ ) of
	0 -> [Str];	    
	I -> [string:substr(Str, 1, I-1) |
	      parse_tokens(string:substr(Str, I+1))]
    end.

snd([_, S | _]) -> S.
