-module(server).

-behavior(gen_server).

%% API
-export([start_link/0]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2,
	 terminate/2, code_change/3]).

%% server function
-export([accept_connection/1, broadcast/4, list/3]). 

%% gen_server records
-record(state, {users, channels}).

%%====================================================================
%% API
%%====================================================================
%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%% Description: Starts the server
%%--------------------------------------------------------------------
start_link() ->
    gen_server:start_link({global, server}, ?MODULE, [], []).

%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%--------------------------------------------------------------------
init([]) ->
    io:format("Launching server..., init~n"),
    case gen_tcp:listen(6667,[{packet, line}, {reuseaddr, true}]) of
	{ok, Lsocket} -> spawn(server, accept_connection,[Lsocket]);
	{error, Reason} -> io:format("Listen problem : ~p~n",[Reason])
    end,
    {ok, #state{users=[], channels=[]}}.

%%--------------------------------------------------------------------
%% Function: %% handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------
handle_call({nick, Pid, Nick}, _From, State) ->
    case lists:keysearch(Nick, 2, State#state.users) of
        {value, {Otherpid, Nick}} -> {reply, fail, State};
	false -> case lists:keysearch(Pid, 1, State#state.users) of
                     {value, {Pid, Oldnick}} -> {reply, ok,
                                                 #state{users=[{Pid, Nick}|lists:delete({Pid, Oldnick}, State#state.users)],
                                                        channels=State#state.channels}};
		     false -> {reply, ok,
                               #state{users=[{Pid, Nick}|State#state.users], channels=State#state.channels}}
                 end
    end;

handle_call({c_pid, Channel}, _From, State) ->
    case lists:keysearch(Channel,2,State#state.channels) of
	{value, {Pid, Channel}} -> {reply, {ok, Pid}, State};
	false -> {reply, fail, State}
    end;

handle_call({u_pid, Nick}, _From, State) ->
   case lists:keysearch(Nick, 2, State#state.users) of
	{value, {Pid, Nick}} -> {reply, {ok, Pid}, State};
	false -> {reply, fail, State}
   end;

handle_call({list, Pid, Nick}, _From, State) ->
    server:list(State#state.channels, Pid, Nick),
    {reply, ok, State};

handle_call(Request, _From, State) ->
    io:format("Call ~p not understood~n", [Request]),
    {reply, fail, State}.
    
%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------
handle_cast({channel, join, Channel, Piduser}, State) ->
    case lists:keysearch(Channel,2,State#state.channels) of
        {value, {Pid, Channel}} -> gen_server:cast(Pid,{join,Piduser}),
				   gen_server:cast(Piduser,{join, Pid}),
				   {noreply, State};
        false -> {ok, Pid} = channel:start_link(erlang:make_ref()),
                 gen_server:call(Pid,{create,Channel}),
		 gen_server:cast(Pid,{join,Piduser}),
		 gen_server:cast(Piduser,{join, Pid}),
                 {noreply, #state{users = State#state.users,
                                  channels = [{Pid, Channel}|State#state.channels]}}
    end;

handle_cast({broadcast, User, Action, Args}, State) ->
    server:broadcast(State#state.users, User, Action, Args),
    {noreply, State};

handle_cast({userexit,User}, State) ->
    {noreply, #state{users = lists:delete(User,State#state.users),
		     channels = State#state.channels}};

handle_cast(Msg, State) ->
    io:format("Cast ~p not understood~n", [Msg]),
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%--------------------------------------------------------------------
handle_info({tcp, Socket, Data}, State) ->
    io:format("tcp message: ~p~n", [Data]),
    server:handle_irc(Socket,irc:parse_cmd(Data)),
    {noreply, State};

handle_info(Info, State) ->
    io:format("Info ~p not understood~n", [Info]),
    {noreply, State}.

%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(_Reason, _State) ->
    io:format("server down~n"),
    ok.

%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------
accept_connection(Lsocket) ->
    io:format("listening ...~n"),
    case gen_tcp:accept(Lsocket) of
        {ok, Socket} ->
             io:format("connection~n"),
	     Ref = erlang:make_ref(),
	     {ok, Pid} = irc_user:start_link(Ref),
	     gen_tcp:controlling_process(Socket, Pid),
	     gen_server:cast(Pid,{create,Socket}),
	     server:accept_connection(Lsocket);
        {error, Reason} ->
             io:format("Accept problem : ~p~n",[Reason]),
             server:accept_connection(Lsocket)
    end.

broadcast(Users, User, Action, Args) ->
    case Users of
	[] -> ok;
	[{Pid, _}|Tail] -> gen_server:cast(Pid,{sendexp,User,Action, Args}),
			      server:broadcast(Tail, User, Action, Args)
    end.

list(Channels, Pid, Nick) ->
    case Channels of
	[{C_pid, Channel}|T] -> Topic = gen_server:call(C_pid,{get, topic}),
				gen_server:cast(Pid,{send, "322", [Nick,Channel,Topic]}),
				server:list(T, Pid, Nick);
	[] -> gen_server:cast(Pid,{send, "323", [Nick, "End of LIST"]})
    end.
